@@ -15,12 +15,13 @@ from fs.errors import DestinationExists, DirectoryExists, DirectoryExpected, Dir
 from fs.info import Info
 from fs.iotools import RawWrapper
 from fs.mode import Mode
-from fs.path import basename, dirname, iteratepath, join
+from fs.path import basename, dirname, iteratepath, join, split
 from fs.subfs import SubFS
 from fs.time import datetime_to_epoch, epoch_to_datetime
 
 _fileMimeType = "application/vnd.google-apps.file"
 _folderMimeType = "application/vnd.google-apps.folder"
+_shortcutMimeType = "application/vnd.google-apps.shortcut"
 _sharingUrl = "https://drive.google.com/open?id="
 _INVALID_PATH_CHARS = ":\0"
 _log = getLogger("fs.googledrivefs")
@@ -127,6 +128,11 @@ class SubGoogleDriveFS(SubFS):
 		fs, delegatePath = self.delegate_path(path)
 		fs.remove_parent(delegatePath)
 
+	def add_shortcut(self, shortcut_path, target_path):
+		fs, shortcutPathDelegate = self.delegate_path(shortcut_path)
+		fs, targetPathDelegate = self.delegate_path(target_path)
+		fs.add_shortcut(shortcutPathDelegate, targetPathDelegate)
+
 class GoogleDriveFS(FS):
 	subfs_class = SubGoogleDriveFS
 
@@ -135,6 +141,7 @@ class GoogleDriveFS(FS):
 
 		self.drive = build("drive", "v3", credentials=credentials, cache_discovery=False)
 		self.retryCount = 3
+		self.enforceSingleParent = False
 
 		_meta = self._meta = {
 			"case_insensitive": True,
@@ -213,6 +220,7 @@ class GoogleDriveFS(FS):
 		isRoot = isinstance(metadata, list) or metadata == _rootMetadata
 		isFolder = isRoot or (metadata["mimeType"] == _folderMimeType)
 		rfc3339 = "%Y-%m-%dT%H:%M:%S.%fZ"
+		permissions = metadata.get("permissions", None)
 		rawInfo = {
 			"basic": {
 				"name": "" if isRoot else metadata["name"],
@@ -228,8 +236,8 @@ class GoogleDriveFS(FS):
 			},
 			"sharing": {
 				"id": None if isRoot else metadata["id"],
-				"permissions": None if isRoot else metadata["permissions"],
-				"is_shared": None if isRoot else len(metadata["permissions"]) > 1
+				"permissions": permissions,
+				"is_shared": len(permissions) > 1 if permissions is not None else None
 			}
 		}
 		if "contentHints" in metadata and "indexableText" in metadata["contentHints"]:
@@ -316,7 +324,7 @@ class GoogleDriveFS(FS):
 			return [x.name for x in self.scandir(path)]
 
 	def _createSubdirectory(self, name, path, parents):
-		newMetadata = {"name": basename(name), "parents": parents, "mimeType": _folderMimeType}
+		newMetadata = {"name": basename(name), "parents": parents, "mimeType": _folderMimeType, "enforceSingleParent": self.enforceSingleParent}
 		self.drive.files().create(body=newMetadata, fields="id").execute(num_retries=self.retryCount)
 		return SubGoogleDriveFS(self, path)
 
@@ -435,7 +443,7 @@ class GoogleDriveFS(FS):
 			if dstItem is not None:
 				self.drive.files().delete(fileId=dstItem["id"]).execute(num_retries=self.retryCount)
 
-			newMetadata = {"parents": [parentDirItem["id"]], "name": basename(dst_path)}
+			newMetadata = {"parents": [parentDirItem["id"]], "name": basename(dst_path), "enforceSingleParent": self.enforceSingleParent}
 			self.drive.files().copy(fileId=srcItem["id"], body=newMetadata).execute(num_retries=self.retryCount)
 
 	# Non-essential - takes advantage of the file contents already being on the server
@@ -474,10 +482,11 @@ class GoogleDriveFS(FS):
 				fileId=srcItem["id"],
 				addParents=dstParentDirItem["id"],
 				removeParents=srcParentItem["id"],
-				body={"name": basename(dst_path)}).execute(num_retries=self.retryCount)
+				body={"name": basename(dst_path), "enforceSingleParent": self.enforceSingleParent}).execute(num_retries=self.retryCount)
 
 	def add_parent(self, path, parent_dir):
 		_log.info(f"add_parent: {path} -> {parent_dir}")
+		_log.warning("Multiple parents feature is expected to be removed on 2020-09-30")
 		_CheckPath(path)
 		_CheckPath(parent_dir)
 		with self._lock:
@@ -502,10 +511,11 @@ class GoogleDriveFS(FS):
 			self.drive.files().update(
 				fileId=sourceItem["id"],
 				addParents=parentDirItem["id"],
-				body={}).execute(num_retries=self.retryCount)
+				body={"enforceSingleParent": self.enforceSingleParent}).execute(num_retries=self.retryCount)
 
 	def remove_parent(self, path):
 		_log.info(f"remove_parent: {path}")
+		_log.warning("Multiple parents feature is expected to be removed on 2020-09-30")
 		_CheckPath(path)
 		with self._lock:
 			idsFromPath = self._itemsFromPath(path)
@@ -515,4 +525,39 @@ class GoogleDriveFS(FS):
 			self.drive.files().update(
 				fileId=sourceItem["id"],
 				removeParents=idsFromPath[dirname(path)]["id"],
-				body={}).execute(num_retries=self.retryCount)
+				body={"enforceSingleParent": self.enforceSingleParent}).execute(num_retries=self.retryCount)
+
+	def add_shortcut(self, shortcut_path, target_path):
+		_log.info(f"add_shortcut: {shortcut_path}, {target_path}")
+		_CheckPath(shortcut_path)
+		_CheckPath(target_path)
+
+		with self._lock:
+			idsFromTargetPath = self._itemsFromPath(target_path)
+			if target_path not in idsFromTargetPath:
+				raise ResourceNotFound(path=target_path)
+
+			targetItem = idsFromTargetPath[target_path]
+			if targetItem["mimeType"] == _folderMimeType:
+				raise FileExpected(target_path)
+
+			idsFromShortcutPath = self._itemsFromPath(shortcut_path)
+			if shortcut_path in idsFromShortcutPath:
+				raise DestinationExists(shortcut_path)
+
+			shortcutParentDir, shortcutName = split(shortcut_path)
+			shortcutParentDirItem = idsFromShortcutPath.get(shortcutParentDir)
+			if shortcutParentDirItem is None:
+				raise ResourceNotFound(shortcutParentDir)
+
+			metadata = {
+				"name": shortcutName,
+				"parents": [shortcutParentDirItem["id"]],
+				"mimeType": _shortcutMimeType,
+				"shortcutDetails": {
+					"targetId": targetItem["id"]
+				},
+				"enforceSingleParent": self.enforceSingleParent
+			}
+
+			_ = self.drive.files().create(body=metadata, fields="id").execute(num_retries=self.retryCount)
